/**
 * @file
 *
 * Copyright (c) 2010 Aalto University) and RWTH Aachen University.
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 *
 * This file provides a framework for modifying HIP packets. It includes
 * adding new parameters in the correct order and adapting the various
 * headers.
 *
 * @brief Framework for the midauth extensions
 *
 * @author Thomas Jansen
 * @author Dominic Gatzen
 *
 * @note: According to draft-heer-hip-middle-auth-00 we SHOULD support IP-level
 * fragmentation for IPv6 and MUST support IP-level fragmentation for IPv4.
 * Currently we do neither.
 */

#define _BSD_SOURCE

#include <stdint.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <linux/netfilter.h>

#include "lib/core/builder.h"
#include "lib/core/debug.h"
#include "lib/core/ife.h"
#include "lib/core/protodefs.h"
#include "lib/core/solve.h"
#include "lib/tool/checksum.h"
#include "firewall_defines.h"
#include "pisa.h"
#include "midauth.h"


static struct midauth_handlers handlers;

/**
 * Changes IPv4 header to match new length and updates the checksum.
 *
 * @param ip  a pointer to the IPv4 header
 * @param len new payload length
 */
static void update_ipv4_header(struct iphdr *ip, int len)
{
    unsigned short *w = (unsigned short *) ip;
    int hdrlen, checksum = 0;

    ip->tot_len = htons(len);
    ip->check   = 0;

    for (hdrlen = ip->ihl * 4, checksum = 0; hdrlen > 1; hdrlen -= 2) {
        checksum += *w++;
    }
    if (hdrlen == 1) {
        unsigned short padding = 0;
        *(unsigned char *) (&padding) = *(unsigned char *) w;
        checksum                     += padding;
    }

    checksum  = (checksum >> 16) + (checksum & 0xffff);
    checksum += (checksum >> 16);

    ip->check = ~checksum;
}

/**
 * Changes IPv6 header to match new length.
 *
 * @param ip a pointer to the IPv6 header
 * @param len new IPv6 packet length
 */
static void update_ipv6_header(struct ip6_hdr *ip, int len)
{
    ip->ip6_ctlun.ip6_un1.ip6_un1_plen = htons(len - sizeof(struct ip6_hdr));
}

#define CHECKSUM_CARRY(x) \
    (x = (x >> 16) + (x & 0xffff), (~(x + (x >> 16)) & 0xffff))

/**
 * Changes UDP header for IPv4 to match new content and updates the checksum.
 *
 * @param ip a pointer to the IPv4 header, not the UDP header
 * @param len total length of the IPv4 packet
 */
static void update_udp_header(struct iphdr *ip, int len)
{
    unsigned long sum;
    uint16_t *w       = (uint16_t *) ((unsigned char *) ip + (ip->ihl * 4));
    uint16_t protocol = ntohs(IPPROTO_UDP);
    int i;
    struct udphdr *udp = (struct udphdr *) w;

    len -= ip->ihl * 4;

    udp->check = 0;
    udp->len = htons(len);

    /* UDP header and data */
    sum = 0;
    while (len > 0) {
        sum += *w++;
        len -= 2;
    }
    if (len == 1) {
        unsigned short padding = 0;
        *(unsigned char *) (&padding) = *(unsigned char *) w;
        sum += padding;
    }

    /* add UDP pseudoheader */
    w          = (uint16_t *) &ip->saddr;
    for (i = 0; i < 4; w++, i++) {
        sum += *w;
    }
    sum += protocol;
    sum += udp->len;

    /* set the checksum */
    udp->check = (CHECKSUM_CARRY(sum));
}

/**
 * Calculate the new checksum for the HIP packet in IPv4. Note that UDP
 * encapsulated HIP packets don't have a checksum. Therefore don't call this
 * function for them.
 *
 * @param ip the modified IP packet
 */
static void update_hip_checksum_ipv4(struct iphdr *ip)
{
    struct sockaddr_in src, dst;
    struct hip_common *msg = (struct hip_common *) ((char *) ip +
                                                    (ip->ihl * 4));

    memset(&src, 0, sizeof(src));
    memset(&dst, 0, sizeof(dst));

    src.sin_family = AF_INET;
    memcpy(&src.sin_addr, &ip->saddr, sizeof(uint32_t));

    dst.sin_family = AF_INET;
    memcpy(&dst.sin_addr, &ip->daddr, sizeof(uint32_t));

    hip_zero_msg_checksum(msg);
    msg->checksum  = hip_checksum_packet((char *) msg,
                                         (struct sockaddr *) &src,
                                         (struct sockaddr *) &dst);
}

/**
 * Calculate the new checksum for the HIP packet in IPv6.
 *
 * @param ip the modified IP packet
 */
static void update_hip_checksum_ipv6(struct ip6_hdr *ip)
{
    struct sockaddr_in6 src, dst;
    struct hip_common *msg = (struct hip_common *) ((char *) ip +
                                                    sizeof(struct ip6_hdr));

    memset(&src, 0, sizeof(src));
    memset(&dst, 0, sizeof(dst));

    src.sin6_family = AF_INET6;
    memcpy(&src.sin6_addr, &ip->ip6_src, sizeof(struct in6_addr));

    dst.sin6_family = AF_INET6;
    memcpy(&dst.sin6_addr, &ip->ip6_dst, sizeof(struct in6_addr));

    hip_zero_msg_checksum(msg);
    msg->checksum   = hip_checksum_packet((char *) msg,
                                          (struct sockaddr *) &src,
                                          (struct sockaddr *) &dst);
}

/**
 * Take care of adapting all headers in front of the HIP payload to the new
 * content.
 *
 * @param ctx context of the modified midauth packet
 */
static void midauth_update_all_headers(hip_fw_context_t *ctx)
{
    struct iphdr *ipv4   = NULL;
    struct ip6_hdr *ipv6 = NULL;
    size_t len           = 0;

    len = hip_get_msg_total_len(ctx->transport_hdr.hip);

    switch (ctx->ip_version) {
    case 4:
        ipv4 = (struct iphdr *) ctx->ipq_packet->payload;
        len += ipv4->ihl * 4;
        if (ipv4->protocol == IPPROTO_UDP) {
            len += sizeof(struct udphdr) + HIP_UDP_ZERO_BYTES_LEN;
            update_udp_header(ipv4, len);
        } else {
            update_hip_checksum_ipv4(ipv4);
        }
        update_ipv4_header(ipv4, len);
        break;
    case 6:
        ipv6 = (struct ip6_hdr *) ctx->ipq_packet->payload;
        len += sizeof(struct ip6_hdr);
        update_hip_checksum_ipv6(ipv6);
        update_ipv6_header(ipv6, len);
        break;
    default:
        HIP_ERROR("Unknown IP version. %i, expected 4 or 6.\n",
                  ctx->ip_version);
        break;
    }

    ctx->ipq_packet->data_len = len;
}

/**
 * Verify that the challenge response in a packet is valid
 *
 * @param hip packet that contains the challenge response
 * @param s   challenge response parameter
 * @return    0 on success, <0 otherwise
 */
int midauth_verify_challenge_response(struct hip_common *hip,
                                      struct hip_challenge_response *s)
{
    int err = 0;
    struct hip_solution solution;
    uint8_t digist[HIP_AH_SHA_LEN];

    HIP_IFEL(hip_build_digest(HIP_DIGEST_SHA1, s->opaque, 24, digist) < 0,
             -1, "Building of SHA1 Random seed I failed\n");
    solution.K        = s->K;
    solution.reserved = 0;
    solution.I        = *digist & 0x40;
    solution.J        = s->J;
    HIP_DEBUG("solution: %d \n", solution.J);

    HIP_IFEL(hip_solve_puzzle(&solution, hip, HIP_VERIFY_PUZZLE) == 0,
             -1, "Solution is wrong\n");

out_err:
    return err;
}

/**
 * Move the last HIP parameter to the correct position according to its
 * parameter type. Will probably break the packet if something is moved in
 * front of a signature.
 *
 * @param hip the HIP packet
 * @return 0 on success
 */
static int midauth_relocate_last_hip_parameter(struct hip_common *hip)
{
    int err                  = 0, len, total_len, offset;
    char buffer[HIP_MAX_PACKET], *ptr = (char *) hip;
    struct hip_tlv_common *i = NULL, *last = NULL;
    hip_tlv_type_t type;

    while ((i = hip_get_next_param_readwrite(hip, i))) {
        last = i;
    }

    HIP_IFEL(last == NULL, -1, "Trying to relocate in an empty packet!\n");

    total_len = hip_get_msg_total_len(hip);
    len       = hip_get_param_total_len(last);
    type      = hip_get_param_type(last);

    HIP_IFEL(len > (int)sizeof(buffer), -1,
             "Last parameter's length exceeds HIP_MAX_PACKET\n");

    /* @todo check for signature parameter to avoid broken packets */

    memcpy(buffer, last, len);
    i = NULL;

    while ((i = hip_get_next_param_readwrite(hip, i))) {
        if (hip_get_param_type(i) > type) {
            offset = (char *) i - (char *) hip;

            memmove(ptr + offset + len, ptr + offset,
                    total_len - offset - len);
            memcpy(ptr + offset, buffer, len);
            break;
        }
    }

out_err:
    return err;
}

/**
 * Creates a challenge request and adds it to a forwarded HIP
 * packet.
 *
 * @param ctx          connection context of the modified packet
 * @param val_K        puzzle difficulty
 * @param ltime        lifetime of the challenge in s
 * @param opaque       contents of the opaque data field
 * @param opaque_len   length of the opaque data field
 * @return 0 on success, <0 otherwise
 */
int midauth_add_challenge_request(hip_fw_context_t *ctx, uint8_t val_K,
                                  uint8_t ltime,
                                  uint8_t *opaque,
                                  uint8_t opaque_len)
{
    struct hip_common *hip = ctx->transport_hdr.hip;
    int err                = 0;

    ctx->modified = 1;

    HIP_IFEL(hip_build_param_challenge_request(hip, val_K, ltime,
                                               opaque, opaque_len),
             -1, "Failed to build challenge_request parameter\n");
    HIP_IFEL(midauth_relocate_last_hip_parameter(hip), -1,
             "Failed to relocate new challenge_request parameter\n");

out_err:
    return err;
}

int midauth_handler_accept(UNUSED hip_fw_context_t *ctx)
{
    return NF_ACCEPT;
}

/**
 * Drops a packet. Used in midauth_handlers as a default handler.
 *
 * @param ctx context of the packet
 * @return NF_DROP
 */
static int midauth_handler_drop(UNUSED hip_fw_context_t *ctx)
{
    return NF_DROP;
}

/**
 * Distinguish the different UPDATE packets.
 *
 * @param ctx context of the modified packet
 * @return the verdict, either NF_ACCEPT or NF_DROP
 */
static midauth_handler filter_midauth_update(const hip_fw_context_t *ctx)
{
    if (hip_get_param(ctx->transport_hdr.hip, HIP_PARAM_LOCATOR)) {
        return handlers.u1;
    }
    if (hip_get_param(ctx->transport_hdr.hip, HIP_PARAM_ECHO_REQUEST)) {
        return handlers.u2;
    }
    if (hip_get_param(ctx->transport_hdr.hip, HIP_PARAM_ECHO_RESPONSE)) {
        return handlers.u3;
    }

    HIP_ERROR("Unknown UPDATE format, rejecting the request!\n");
    return midauth_handler_drop;
}

/**
 * Packet handler dispatcher function. Classifies packets based on
 * their type and calls the appropriate type-specific handler functions.
 *
 * @param ctx HIP connection context
 * @return the verdict, either NF_ACCEPT or NF_DROP
 */
int midauth_filter_hip(hip_fw_context_t *ctx)
{
    int verdict               = NF_ACCEPT;
    midauth_handler h         = NULL;
    midauth_handler h_default = midauth_handler_accept;
    /* @todo change this default value to midauth_handler_drop to
     * disallow unknown message types */

    switch (ctx->transport_hdr.hip->type_hdr) {
    case HIP_I1:
        h = handlers.i1;
        break;
    case HIP_R1:
        h = handlers.r1;
        break;
    case HIP_I2:
        h = handlers.i2;
        break;
    case HIP_R2:
        h = handlers.r2;
        break;
    case HIP_UPDATE:
        h = filter_midauth_update(ctx);
        break;
    case HIP_CLOSE:
        h = handlers.close;
        break;
    case HIP_CLOSE_ACK:
        h = handlers.close_ack;
        break;
    default:
        HIP_DEBUG("filtering default message type\n");
        break;
    }

    if (!h) {
        h = h_default;
    }
    verdict = h(ctx);

    /* do not change packet when it is dropped */
    if (verdict != NF_ACCEPT) {
        ctx->modified = 0;
    }

    /* if packet was modified correct every necessary part */
    if (ctx->modified != 0) {
        midauth_update_all_headers(ctx);
    }

    return verdict;
}

/**
 * Call the initializer functions
 */
void midauth_init(void)
{
    pisa_init(&handlers);
}
